#include "PointCloud.h"
#include "../common/Color.h"
#include "../common/oldbuffer.h"
#include "../common/misc.h"
#include "../Primitives/Triangle.h"

extern SlimScene* main_scene;
extern void (*update_frame_output)(SlimScene *scene);

void PointCloud::render(SlimScene *scene)
{
	// rasterizer test (no clipping yet)
	//GraphicMatrix proj = main_scene->camera->getProjection();
	GraphicMatrix proj = main_scene->camera->getProjectionFOV();
	//GraphicMatrix proj = main_scene->camera->getFrustum();
	GraphicMatrix look = main_scene->camera->getLookAt();
	GraphicMatrix transform =  proj * look;
	
	clearCanvas();
	
	for(unsigned int i= 0; i<main_scene->primitives.size(); i++)
	{
		Triangle *t = dynamic_cast<Triangle*>(main_scene->primitives[i]);
		if( t == NULL)
			continue;
		vector4 v4[3];
		for(int i=0; i<3; i++)
		{
			v4[i] = v4fromv3(t->vertex[0]);
			v4[i] = transform * v4[i];
		}
		
		//perspective scaling
		for(int i=0; i<3; i++)
		{
			slimFloat perspectiveScale1 = v4[i][3] == 0.0f ? 1.0 : 1.0f / v4[i][3];
			v4[i] *= perspectiveScale1;
		}
		
		//viewport transform
		vector3 v3[3];
		for(int i=0; i<3; i++)
		{
			v3[i] = viewportTransform(v4[i]);
		}
		
		const int pointSize = 1;
		for(int x=0; x<pointSize; x++)
		{
			for(int y=0; y<pointSize; y++)
			{
				for(int i=0; i<3; i++)
				{
					Color c = t->diff * 255.0f;
					set_Color(v3[i][0]+x, v3[i][1]+y, &c);
					//writePixel(v3[i][0]+x, v3[i][1]+y, t, t->vertex[i], *scene);
				}
			}
		}
		
	}
}
