#ifndef __RASTERIZER
#define __RASTERIZER

#include "RenderEngine.h"
#include "../Primitives/Triangle.h"

void rasterize();

class Rasterizer : public RenderEngine
{
public:
	Rasterizer(Canvas *c) : RenderEngine(c)
	{
		//alloc zbuffer
	}
	
	~Rasterizer()
	{
	}
	
	void render(SlimScene *scene);
	
protected:

	void drawLine(vector3 &p1, vector3 &p2, Color &c);
	
	inline vector3 viewportTransform(vector4 &clippedVertex)
	{
		vector3 viewVertex;
		//viewport transform
		slimFloat w = main_scene->width;
		slimFloat h = main_scene->height;
		slimFloat offsetX = 0;
		slimFloat offsetY = 0;
		viewVertex[0] = (clippedVertex[0] + 1) * w/2 + offsetX;
		viewVertex[1] = (clippedVertex[1] + 1) * h/2 + offsetY;
		viewVertex[1] = h - viewVertex[1]; //invert opengl y-coord
		
		return viewVertex;
	}
	
	inline void writePixel(int x, int y, PrimitiveBase *primitive, vector3 &vertex, SlimScene &scene)
	{
		intersect_data i_data;
		Color c;
		float distance;
		vector3 proj;
		
		proj = vertex - scene.camera->pos;
		distance = proj.length();
		if(distance > scene.hitData->getIndex(x, y).distance)
			return;
		
		scene.hitData->getIndex(x, y).distance = distance;
		i_data.obj = primitive;
		i_data.proj = proj.normalize();
		i_data.intersect = vertex;
		i_data.obj->shader(&i_data, &c);
		
		set_Color(x, y, &c);
	}
	
};

#endif
