#include "../../common/oldbuffer.h"
#include "../../common/Color.h"
#include "../shaders.h"
#include "../../parse/parser.h"
#include "../../frontend/globals.h"
#include <stdlib.h>
#ifdef USE_OMP
#include <omp.h>
#endif

typedef struct {
	Camera previousCamera;
	Buffer<Color> *colorBuffer;
	Buffer<HitPoint> *hitData;
} reproj_data;


int _init_reproj(void **data_ptr, char *input)
{
	reproj_data *rpd = (reproj_data*) malloc(sizeof(reproj_data));
	rpd->previousCamera = *main_scene->camera;
	rpd->colorBuffer = new Buffer<Color>(main_scene->width, main_scene->height);
	rpd->hitData = new Buffer<HitPoint>(main_scene->width, main_scene->height);
	
	*data_ptr = (void*)rpd;
	return 1;
}

void reprojectSample(Color *oldColor, vector3 &hitPoint)
{
	vector3 newProjection = hitPoint - main_scene->camera->pos;
	vector3 newCoords;
	
	main_scene->camera->getImageCoords(newCoords, newProjection);
	newCoords[0] = roundf(newCoords[0]);
	newCoords[1] = roundf(newCoords[1]);
	if(newCoords.c[0] < 0 || newCoords.c[0] > main_scene->width ||
	   newCoords.c[1] < 0 || newCoords.c[1] > main_scene->height)
		return;
	//if(oldIntersection.normal.dot(newProjection) > 0)
		//return;
	if(main_scene->camera->look.dot(newProjection) < 0)
		return;
	
	//float distance = newProjection.length();
	//if(main_scene->hitData->getIndex(newCoords.c[0], newCoords.c[1]).distance > distance)
	//	main_scene->hitData->getIndex(newCoords.c[0], newCoords.c[1]).distance = distance;
	//else
	//	return;
	
	set_Color(newCoords.c[0], newCoords.c[1], oldColor);
}

void _run_reproj(void *data)
{
	reproj_data *rpd = (reproj_data*)data;
	int x,y;
	intersect_data old_i;
	intersect_data *new_i;
	Color new_c;
	Color old_c;
	static char filled = 0;
	
	//copy the framebuffer colors
	if(filled == 0) {
		rpd->colorBuffer->copy(*main_scene->colorOutput);
		rpd->hitData->copy(*main_scene->hitData);
	}

	//turn off rendering
	if(filled == 0)
		process_command_string("engine null");
	
	filled = 1;
		
	//reprojection time
	//clear framebuffer
	//memset(main_scene->Color_front, 0, sizeof(Color)*main_scene->width*main_scene->height);
	main_scene->colorOutput->reset();
	main_scene->hitData->reset();
	
	//#pragma omp parallel for schedule(dynamic, 10)
	for(y = 0; y < main_scene->height; y++)
	{
		for(x = 0; x < main_scene->width; x++)
		{
			//copy_hit_data(&old_i, get_idata(x, y) );
			old_i.intersect = main_scene->hitData->getIndex(x,y).pos;

			//get the old Color
			old_c = rpd->colorBuffer->getIndex(x,y);

			//if(old_i.obj_num > 0) {
			if(old_i.obj != main_scene->background)
			{
				reprojectSample(&old_c, old_i.intersect);
			}
		}
	}
	
	rpd->previousCamera = *main_scene->camera;
}


void _cleanup_reproj(void* data)
{
	free((reproj_data*)data);
	process_command_string("engine normal");
}

shader_data reproj =
{
	"reproj",
	"options:\n   reproj <float alpha>\n",
	(int (*)(void*, char*))_init_reproj,
	(void* (*)(void*))_run_reproj,
	(void* (*)(void*))_cleanup_reproj,
};

shader_data* get_reproj()
{
	return &reproj;
}
